/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.lang.management.ManagementFactory;
import java.rmi.NoSuchObjectException;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Hashtable;
import java.util.Properties;
import java.util.logging.Level;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;

import org.ow2.orchestra.cluster.JmxServer;
import org.ow2.orchestra.cluster.Server;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.jmx.RemoteDeployer;
import org.ow2.orchestra.facade.jmx.RemoteDeployerMBean;
import org.ow2.orchestra.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.pvm.internal.jobexecutor.AbstractJobExecutor;
import org.ow2.orchestra.pvm.internal.jobexecutor.JobExecutor;
import org.ow2.orchestra.pvm.internal.jobexecutor.pvm.PvmJobExecutor;
import org.ow2.orchestra.services.InvokeExecutor;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.OrchestraConstants;

public class JMXAgent {

  private MBeanServer mbs;

  private ObjectName apiObjectName;
  private ObjectName invokeExecutorObjectName;
  private ObjectName jobExecutorObjectName;

  private JMXConnectorServer cs;

  private Registry registry;

  private EnvironmentFactory environmentFactory;

  public void startOrchestra(final EnvironmentFactory envFactory) {

    this.environmentFactory = envFactory;

    // Orchestra properties
    final Properties orchestraProperties = (Properties) envFactory.get(OrchestraConstants.ORCHESTRA_PROPERTIES);
    final String jmxServiceUrl = orchestraProperties.getProperty(OrchestraConstants.JMX_SERVICE_URL);
    final String jmxObjectName = orchestraProperties.getProperty(OrchestraConstants.JMX_OBJECT_NAME);

    final Server orchestraServer = new JmxServer(jmxServiceUrl, jmxObjectName);
    // create api object
    final RemoteDeployerMBean remoteDeployer = new RemoteDeployer(orchestraServer, envFactory);

    // reinitialize orchestra (deploys web services)
    remoteDeployer.initialize();

    // start job executor after web services are deployed
    final JobExecutor jobExecutor = envFactory.get(JobExecutor.class);
    if (jobExecutor != null) {
      if (jobExecutor instanceof AbstractJobExecutor) {
        ((AbstractJobExecutor)jobExecutor).setName(orchestraServer.toString());
      }
      jobExecutor.start();
    }

    // expose api as MBean
    try {
      // start rmiregistry
      final int jmxPort = Integer.parseInt(orchestraProperties.getProperty(OrchestraConstants.JMX_PORT));
      this.registry = java.rmi.registry.LocateRegistry.createRegistry(jmxPort);


      this.mbs = ManagementFactory.getPlatformMBeanServer();

      // Uniquely identify the MBeans and register them with the MBeanServer
      this.apiObjectName = ObjectName.getInstance(jmxObjectName);
      this.mbs.registerMBean(remoteDeployer, this.apiObjectName);

      final InvokeExecutor invokeExecutor = envFactory.get(InvokeExecutor.class);
      if (invokeExecutor != null) {
        final Hashtable<String,String> props = new Hashtable<String, String>();
        props.put("type", "threadPool");
        props.put("name", "InvokeExecutor");
        this.invokeExecutorObjectName =
          ObjectName.getInstance(this.apiObjectName.getDomain(), props);
        this.mbs.registerMBean(new ManagedThreadPool(invokeExecutor.getExecutorService()), this.invokeExecutorObjectName);
      }
      if (jobExecutor != null && (jobExecutor instanceof PvmJobExecutor)) {
        final PvmJobExecutor pvmJobExecutor = (PvmJobExecutor) jobExecutor;
        final Hashtable<String,String> props = new Hashtable<String, String>();
        props.put("type", "threadPool");
        props.put("name", "JobExecutor");
        this.jobExecutorObjectName =
          ObjectName.getInstance(this.apiObjectName.getDomain(), props);
        this.mbs.registerMBean(new ManagedThreadPool(pvmJobExecutor.getThreadPool()), this.jobExecutorObjectName);
      }

      final JMXServiceURL url = new JMXServiceURL(jmxServiceUrl);
      this.cs = JMXConnectorServerFactory.newJMXConnectorServer(url, null, this.mbs);
      this.cs.start();

      Misc.log(Level.INFO, "Orchestra is started using JMX connector [" + this.cs.getAddress().toString() + "]");
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Problem registering mbean in mbean server", e);
    }


  }

  public void stopOrchestra() {
    if (this.mbs != null) {
      try {
        this.cs.stop();
        if (this.apiObjectName != null) {
          this.mbs.unregisterMBean(this.apiObjectName);
        }
        if (this.invokeExecutorObjectName != null) {
          this.mbs.unregisterMBean(this.invokeExecutorObjectName);
        }
        if (this.jobExecutorObjectName != null) {
          this.mbs.unregisterMBean(this.jobExecutorObjectName);
        }
        this.mbs = null;
      } catch (final Exception e) {
        throw new OrchestraRuntimeException("Problem unregistering mbean", e);
      }
    }
    if (this.registry != null) {
      try {
        UnicastRemoteObject.unexportObject(this.registry, true);
      } catch (final NoSuchObjectException e) {
        throw new OrchestraRuntimeException("Problem unregistering rmi registry", e);
      }
      this.registry = null;
    }
    if (this.environmentFactory != null) {
      final JobExecutor jobExecutor = this.environmentFactory.get(JobExecutor.class);
      if (jobExecutor != null) {
        jobExecutor.stop(true);
      }
      final InvokeExecutor invokeExecutor = this.environmentFactory.get(InvokeExecutor.class);
      if (invokeExecutor != null) {
        invokeExecutor.getExecutorService().shutdown();
      }
      this.environmentFactory.close();
      this.environmentFactory = null;
    }
  };

}
