package org.ow2.orchestra.persistence.db;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.hibernate.FlushMode;
import org.hibernate.LockMode;
import org.hibernate.LockOptions;
import org.hibernate.Query;
import org.hibernate.ScrollableResults;
import org.hibernate.Session;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.BpelProcess.InstanceCreationLock;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.deployment.ProcessDeploymentCounter;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.persistence.RuntimeDbSession;
import org.ow2.orchestra.pvm.internal.hibernate.HibernatePvmDbSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.WaitingExecution;
import org.ow2.orchestra.runtime.WaitingExecution.WaitingExecutionIterator;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.PendingMessage;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.w3c.dom.Document;

public class RuntimeDbSessionImpl extends
    HibernatePvmDbSession implements RuntimeDbSession {

  private static final Logger LOG = Logger.getLogger(RuntimeDbSessionImpl.class.getName());

  public RuntimeDbSessionImpl(final Session session) {
    this.setSession(session);
    session.setFlushMode(FlushMode.COMMIT);
  }

  public BpelProcess findBpelProcess(final QName processQName, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findLatestBpelProcessForLock");
      query.setString("qname", processQName.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final InstanceCreationLock res = (InstanceCreationLock) query.uniqueResult();
      if (res == null) {
        return null;
      }
    }
    return this.findActiveBpelProcess(processQName);
  }

  public BpelProcess findActiveBpelProcess(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("findLatestBpelProcess");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    query.setCacheable(true);
    return (BpelProcess) query.uniqueResult();
  }

  public List<BpelProcess> findBpelProcesses(final QName processQName, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findBpelProcessesByQNameForLock");
      query.setString("qname", processQName.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final List<InstanceCreationLock> res = query.list();
      if (res == null || res.isEmpty()) {
        return null;
      }
    }
    return this.findBpelProcesses(processQName);
  }

  private List<BpelProcess> findBpelProcesses(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("findBpelProcessesByQName");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    query.setCacheable(true);
    return query.list();
  }

  public BpelProcess findBpelProcess(final ProcessDefinitionUUID pduuid, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findBpelProcessForLock");
      query.setString("uuid", pduuid.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final InstanceCreationLock res = (InstanceCreationLock) query.uniqueResult();
      if (res == null) {
        return null;
      }
    }
    return this.getBpelProcess(pduuid);
  }

  public BpelProcess getBpelProcess(final ProcessDefinitionUUID pduuid) {
    final Query query = this.getSession().getNamedQuery("getBpelProcessByUUID");
    query.setString("uuid", pduuid.toString());
    query.setCacheable(true);
    final BpelProcess processDefinition = (BpelProcess) query.uniqueResult();
    return processDefinition;
  }

  public List<BpelProcess> getBpelProcesses() {
    final Query query = this.getSession().getNamedQuery("getBpelProcesses");
    query.setCacheable(true);
    return query.list();
  }

  private static class WaitingExecutionIteratorImpl implements WaitingExecutionIterator {

    private final ScrollableResults results;
    private final Session session;
    private WaitingExecution current = null;

    public WaitingExecutionIteratorImpl(final ScrollableResults results, final Session session) {
      this.results = results;
      this.session = session;
    }
    public WaitingExecution next() {
      if (this.results == null) {
        return null;
      }
      if (this.current != null) {
        this.session.evict(this.current);
      }
      if (this.results.next()) {
        final Serializable id = (Serializable) this.results.get(0);
        this.current = (WaitingExecution) this.session.get(WaitingExecution.class, id);
        if (this.current == null) {
          Misc.log(Level.INFO, "WaitingExecution %s not found, skipping.", id);
          return this.next();
        }
        return this.current;
      } else {
        this.results.close();
      }
      return null;
    }
  }


  private List<Set<QName>> getPermutations(final Set<QName> firstQNames, final Set<Correlation> optionalCorrelations) {
    if (optionalCorrelations.isEmpty()) {
      return Arrays.asList(firstQNames);
    }
    final Correlation firstAvailable = optionalCorrelations.iterator().next();
    optionalCorrelations.remove(firstAvailable);

    final List<Set<QName>> result = new ArrayList<Set<QName>>();
    final Set<QName> res1  = new HashSet<QName>(firstQNames);
    res1.addAll(firstAvailable.getCorrelationSet().getProperties().keySet());
    result.addAll(this.getPermutations(res1, optionalCorrelations));
    final Set<QName> res2  = new HashSet<QName>(firstQNames);
    result.addAll(this.getPermutations(res2, optionalCorrelations));
    return result;
  }

  public WaitingExecutionIterator getWaitingExecutions(
      final ReceivingElement receivingElement,
      final Map<QName, Document> csProperties) {

    final Set<QName> requiredProperties = new HashSet<QName>();
    final Set<Correlation> optionalCorrelations = new HashSet<Correlation>();
    for (final Correlation correlation : receivingElement.getCorrelations()) {
      if (correlation.getInitiate().equals(Initiate.no)) {
        requiredProperties.addAll(correlation.getCorrelationSet().getProperties().keySet());
      } else if (correlation.getInitiate().equals(Initiate.join)) {
        optionalCorrelations.add(correlation);
      }
    }


    final List<Set<QName>> propertiesValues = this.getPermutations(requiredProperties, optionalCorrelations);

    Misc.log(Level.INFO, "getWaitingExecutions:  %s possible correlations found.", propertiesValues.size());

    for (final Set<QName> propQNames : propertiesValues) {
      Misc.log(Level.INFO, "getWaitingExecutions:  trying to match %s correlation properties.", propQNames.size());
      final Map<QName, Document> csProps = new HashMap<QName, Document>(csProperties);
      csProps.keySet().retainAll(propQNames);

      final int correlationKey = ReceivingElementUtil.getCorrelationPropertiesKey(csProps);

      final Query query = this.getSession().getNamedQuery("getWaitingExecutionsByCorrelationKey");
      query.setParameter("activity", receivingElement.getInboundMessageElement());
      query.setInteger("correlationKey", correlationKey);

      final ScrollableResults waitingExecutions = query.scroll();
      if (waitingExecutions.first()) {
        waitingExecutions.beforeFirst();
        return new WaitingExecutionIteratorImpl(waitingExecutions, this.getSession());
      } else {
        waitingExecutions.close();
      }
    }
    return new WaitingExecutionIteratorImpl(null, this.getSession());
  }

  public WaitingExecution getFirstWaitingExecution(final InboundMessageElement inboundMessageElement) {
    final Query query = this.getSession().getNamedQuery("getWaitingExecutions");
    query.setParameter("activity", inboundMessageElement);
    query.setMaxResults(1);
    final Serializable id = (Serializable) query.uniqueResult();
    if (id != null) {
      return this.get(WaitingExecution.class, id);
    }
    return null;
  }

  public Collection<PendingMessage> getPendingMessages(final OperationKey operationKey) {
    final Query query = this.getSession().getNamedQuery("getPendingMessages");
    query.setString("operationKey", operationKey.toString());
    return query.list();
  }

  public List<BpelExecution> getBpelInstances() {
    final Query query = this.getSession().getNamedQuery("getBpelInstances");
    return query.list();
  }

  public BpelExecution getBpelInstance(final ProcessInstanceUUID processInstanceUUID) {
    final Query query = this.getSession().getNamedQuery("getBpelInstance");
    query.setString("processInstanceUUID", processInstanceUUID.toString());
    final BpelExecution bpelExecution = (BpelExecution) query.uniqueResult();
    return bpelExecution;
  }

  public List<BpelExecution> getBpelInstances(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("getBpelInstancesOfProcess");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    return query.list();
  }

  public List<BpelExecution> getBpelInstances(final ProcessDefinitionUUID pduuid) {
    final Query query = this.getSession().getNamedQuery("getBpelInstancesOfProcessByUUID");
    query.setString("uuid", pduuid.toString());
    return query.list();
  }

  public ProcessDeploymentCounter getProcessDeploymentCounter() {
    return (ProcessDeploymentCounter) this.getSession().createCriteria(ProcessDeploymentCounter.class).uniqueResult();
  }

  public void clearProcessCache() {
    this.getSession().getSessionFactory().getCache().evictDefaultQueryRegion();
    this.getSession().getSessionFactory().getCache().evictEntityRegion(BpelProcess.class);
    Misc.fastDynamicLog(RuntimeDbSessionImpl.LOG, Level.INFO, "BpelProcess Cache cleared");
  }
}
