/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.activity.IAbstractActivity;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.data.runtime.VariableUpdateData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard
 */
public class VariableRuntime {

  protected long dbid;
  protected int dbversion;

  /** variable name. */
  protected String name;

  /** variable value. */
  protected Document xmlValue;

  /** variable definition. */
  protected Variable variableDefinition;

  protected ScopeRuntime scopeRuntime;

  protected VariableRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param variableDefinition
   *            variableDefinition
   * @param execution
   *            execution
   */
  public VariableRuntime(final Variable variableDefinition, final BpelExecution execution,
    final ScopeRuntime scopeRuntime) {
    this.scopeRuntime = scopeRuntime;
    this.name = variableDefinition.getName();
    this.variableDefinition = variableDefinition;
    this.xmlValue = null;
  }

  public void initializeVariable(final BpelExecution execution) {
    final Copy initCopy = this.variableDefinition.getVirtualCopy();
    if (initCopy != null) {
      initCopy.execute(execution);
    }
    final IAbstractActivity activity = (IAbstractActivity) execution.getNode().getBehaviour();
    this.updated(activity.createRuntimeData(execution));
  }

  public Variable getVariableDefinition() {
    return this.variableDefinition;
  }

  public void setVariableDefinition(final Variable variableDefinition) {
    this.variableDefinition = variableDefinition;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public Object getValue() {
    if (this.xmlValue == null) {
      this.xmlValue = XmlUtil.getDocumentFromString(XmlUtil.toString(this.variableDefinition.getDefaultValue()));
    }
    if (this.variableDefinition.getMessageType() != null) {
      return new MessageVariable(this.xmlValue);
    }
    if (this.xmlValue == null) {
      return null;
    }
    return this.xmlValue.getDocumentElement();
  }

  public void setValue(final MessageVariable message) {
    if (this.variableDefinition.getMessageType() == null) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.mismatchedAssignmentFailure,
          "trying to set a messageValue and the variableDefinition is not a messageType : " + this.toString(),
          null);
    } else {
      this.xmlValue = message.getPartsAsDocument();
    }
  }

  public void setValue(final Object value, final boolean keepSrcElementName) {
    if (this.variableDefinition.getMessageType() != null) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.mismatchedAssignmentFailure,
          "trying to set a xmlValue and the variableDefinition is a messageType : " + this.toString(),
          null);
    }
    if (this.xmlValue == null) {
      final Element defaultDocElt = this.variableDefinition.getDefaultValue().getDocumentElement();
      this.xmlValue = XmlUtil.getDocumentWithOneElement(
          new QName(defaultDocElt.getNamespaceURI(), defaultDocElt.getLocalName())).getOwnerDocument();
    }
    BpelXmlUtil.setObjectValue(this.xmlValue.getDocumentElement(), value, keepSrcElementName);
  }

  public void updated(final ActivityInstanceData activityInstanceData) {
    if (Variable.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME.equals(this.name)
        || Variable.OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME.equals(this.name)) {
      //do not log anonymous variables
      return;
    }
    if (this.variableDefinition.getMessageType() == null) {
      EnvTool.getRecorder().recordVariableUpdated(
          new VariableUpdateData(
              this.scopeRuntime.getActivityInstanceUUID(),
              activityInstanceData,
              this.name,
              this.xmlValue.getDocumentElement()));
    } else {
      EnvTool.getRecorder().recordVariableUpdated(
          new VariableUpdateData(
              this.scopeRuntime.getActivityInstanceUUID(),
              activityInstanceData,
              this.name,
              new MessageVariable(this.xmlValue)));
    }
  }

  public long getDbid() {
    return this.dbid;
  }

//  public String toString() {
//    return VariableRuntime.class.getName()
//        + "[dbid: " + dbid
//        + ", name: " + name
//        + ", messageValue:" + messageValue
//        + ", xmlValue: " + xmlValue
//        + ", variableDefinition: " + variableDefinition
//        + "]";
//  }
}
