/**
 * Copyright (C) 2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.util.wsdl.WsdlUtil;
import org.w3c.dom.Element;


/**
 *
 * Utility class to retrieve EPR information.
 *
 * @author Guillaume Porcher
 *
 */
public final class AddressingUtil {

  private AddressingUtil() { }

  public static class AddressingInfo {
    private final Element partnerEndPointReference;
    private final String address;
    private final Definition wsdlDefinition;
    private final QName serviceQName;
    private final String portName;

    private Service service;
    private Port port;

    public AddressingInfo(final Element partnerEndPointReference, final String address, final Definition wsdlDefinition,
        final QName serviceQname, final String portName, final Properties orchestraProperties) {

      this.address = AddressingUtil.replacePropertiesInOrchestraURL(address, orchestraProperties);
      this.wsdlDefinition = wsdlDefinition;
      this.serviceQName = serviceQname;
      this.portName = portName;
      this.partnerEndPointReference = partnerEndPointReference;

      if (wsdlDefinition != null) {
        final Map<QName, Service> serviceMap = wsdlDefinition.getServices();
        this.resolveServiceAndPort(serviceMap.values());
      }

    }

    public Element getPartnerEndPointReference() {
      return this.partnerEndPointReference;
    }
    public String getAddress() {
      return this.address;
    }
    public Definition getWsdlDefinition() {
      return this.wsdlDefinition;
    }
    public QName getServiceQName() {
      return this.serviceQName;
    }
    public String getPortName() {
      return this.portName;
    }
    /**
     * @return the service
     */
    public Service getService() {
      return this.service;
    }
    public Port getPort() {
      return this.port;
    }

    public void resolveServiceAndPort(final Collection<Service> knownServices) {
      for (final Service candidateService : knownServices) {
        if (this.serviceQName != null) {
          if (!this.serviceQName.equals(candidateService.getQName())) {
            continue;
          }
        }
        if (this.portName != null) {
          final Port candidatePort = candidateService.getPort(this.portName);
          if (candidatePort != null) {
            this.service = candidateService;
            this.port = candidatePort;
            return;
          }
        } else {
          final Map<String, Port> servicePorts = candidateService.getPorts();
          for (final Port candidatePort : servicePorts.values()) {
            final List<ExtensibilityElement> elements = candidatePort.getExtensibilityElements();
            if (elements.size() > 0) {
              for (final ExtensibilityElement element : elements) {
                if (element instanceof SOAPAddress) {
                  if (((SOAPAddress) element).getLocationURI().equals(this.address)) {
                    this.service = candidateService;
                    this.port = candidatePort;
                    return;
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  public static AddressingInfo getAddressingInfo(final Element serviceRefElem, final Properties orchestraProperties) {
    final String scheme = XmlUtil.attribute(serviceRefElem, "reference-scheme");
    final org.w3c.dom.Element endpointRef = XmlUtil.element(serviceRefElem,
        XmlConstants.ADDRESSING_2004_REF_NAMESPACE, XmlConstants.EPR_TAG_NAME);
    if ((scheme == null || scheme.equals(XmlConstants.ADDRESSING_2004_REF_NAMESPACE))
        && endpointRef != null) {
      return AddressingUtil.getAddressing2004Info(endpointRef, orchestraProperties);
    }
    final org.w3c.dom.Element endpointRef2005 = XmlUtil.element(serviceRefElem,
        XmlConstants.ADDRESSING_2005_REF_NAMESPACE, XmlConstants.EPR_TAG_NAME);
    if ((scheme == null || scheme.equals(XmlConstants.ADDRESSING_2005_REF_NAMESPACE))
        && endpointRef2005 != null) {
      return AddressingUtil.getAddressing2005Info(endpointRef2005, orchestraProperties);
    }
    throw BpelFaultException.generateBpelFaultException(BpelFault.unsupportedReference);
  }

  private static AddressingInfo getAddressing2004Info(final Element endpointRef, final Properties orchestraProperties) {
    final org.w3c.dom.Element addressElt = XmlUtil.element(endpointRef,
        XmlConstants.ADDRESSING_2004_REF_NAMESPACE, XmlConstants.ADDRESS_TAG_NAME);
    final String address;
    if (addressElt != null) {
      address = addressElt.getTextContent();
    } else {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.unsupportedReference, "EndpointReference is missing required 'Address' element.", null);
    }
    QName serviceQname = null;
    String portName = null;
    final org.w3c.dom.Element serviceNameElt = XmlUtil.element(endpointRef,
        XmlConstants.ADDRESSING_2004_REF_NAMESPACE, XmlConstants.SERVICE_NAME_TAG_NAME);
    if (serviceNameElt != null) {
      final String serviceNameText = serviceNameElt.getTextContent();
      final String[] splitted = serviceNameText.split(":", 2);

      if (splitted.length == 2) {
        // if has prefix, resolve namespace
        final String prefix = splitted[0];
        final String namespace = serviceNameElt.lookupNamespaceURI(prefix);
        final String localName = splitted[1];
        serviceQname = new QName(namespace, localName, prefix);
      } else {
        // no prefix, use tns
        serviceQname = new QName(serviceNameElt.lookupNamespaceURI(null), serviceNameText);
      }

      portName = XmlUtil.attribute(serviceNameElt, "PortName");
    }
    return new AddressingInfo(endpointRef, address, null, serviceQname, portName, orchestraProperties);
  }


  private static AddressingInfo getAddressing2005Info(final Element endpointRef, final Properties orchestraProperties) {
    final org.w3c.dom.Element addressElt = XmlUtil.element(endpointRef,
        XmlConstants.ADDRESSING_2005_REF_NAMESPACE, XmlConstants.ADDRESS_TAG_NAME);
    final String address;
    if (addressElt != null) {
      address = addressElt.getTextContent();
    } else {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.unsupportedReference, "EndpointReference is missing required 'Address' element.", null);
    }

    QName serviceQname = null;
    String portName = null;
    Definition wsdlDefinition = null;
    final org.w3c.dom.Element metadataElt = XmlUtil.element(endpointRef,
        XmlConstants.ADDRESSING_2005_REF_NAMESPACE, XmlConstants.METADATA_TAG_NAME);
    if (metadataElt != null) {
      final org.w3c.dom.Element serviceNameElt = XmlUtil.element(endpointRef,
          XmlConstants.ADDRESSING_2005_WSDL_NAMESPACE, XmlConstants.SERVICE_NAME_TAG_NAME);
      if (serviceNameElt != null) {
        final String serviceNameText = serviceNameElt.getTextContent();
        final String[] splitted = serviceNameText.split(":", 2);

        if (splitted.length == 2) {
          // if has prefix, resolve namespace
          final String prefix = splitted[0];
          final String namespace = serviceNameElt.lookupNamespaceURI(prefix);
          final String localName = splitted[1];
          serviceQname = new QName(namespace, localName, prefix);
        } else {
          // no prefix, use tns
          serviceQname = new QName(serviceNameElt.lookupNamespaceURI(null), serviceNameText);
        }

        portName = XmlUtil.attribute(serviceNameElt, "EndPointName");
      }

      final org.w3c.dom.Element wsdlDefinitionElt = XmlUtil.element(endpointRef,
          XmlConstants.XMLNS_WSDL, "definitions");
      if (wsdlDefinitionElt != null) {
        wsdlDefinition = WsdlUtil.readWsdl(wsdlDefinitionElt);
      }
    }

    return new AddressingInfo(endpointRef, address, wsdlDefinition, serviceQname, portName, orchestraProperties);
  }


  private static String getOrchestraURLPrefix(final Properties orchestraProperties) {
    return "http://"
      + orchestraProperties.getProperty(OrchestraConstants.SERVLET_HOST)
      + ":"
      + orchestraProperties.getProperty(OrchestraConstants.SERVLET_PORT)
      + "/"
      + orchestraProperties.getProperty(OrchestraConstants.SERVLET_PATH);
  }

  public static String replacePropertiesInOrchestraURL(final String url, final Properties orchestraProperties) {
    if (url == null) {
      return url;
    }
    // replace ORCHESTRA_WS_URL_KEY by url prefix
    final String url1 = url.replace(
        "${" + OrchestraConstants.ORCHESTRA_WS_URL_KEY + "}",
        AddressingUtil.getOrchestraURLPrefix(orchestraProperties));
    // replace HTTP_PORT_WS_URL_KEY
    final String url2 = url1.replace(
        "${" + OrchestraConstants.HTTP_PORT_WS_URL_KEY + "}",
        orchestraProperties.getProperty(OrchestraConstants.SERVLET_PORT));
    // replace SERVLET_PORT
    final String url3 = url2.replace(
        "${" + OrchestraConstants.SERVLET_PORT + "}",
        orchestraProperties.getProperty(OrchestraConstants.SERVLET_PORT));
    return url3;
  }
}
